@@ -27,8 +27,6 @@
 #include "pangocairo-private.h"
 #include "pango-impl-utils.h"
 
-PangoCairoWarningHistory _pango_cairo_warning_history = { FALSE };
-
 #define PANGO_CAIRO_FONT_PRIVATE(font)		\
   ((PangoCairoFontPrivate *)			\
    (font == NULL ? NULL :			\
@@ -110,8 +108,6 @@ _pango_cairo_font_private_get_scaled_font (PangoCairoFontPrivate *cf_priv)
   cairo_font_face_destroy (font_face);
 
 done:
-  _pango_cairo_font_private_scaled_font_data_destroy (cf_priv->data);
-  cf_priv->data = NULL;
 
   if (G_UNLIKELY (cf_priv->scaled_font == NULL || cairo_scaled_font_status (cf_priv->scaled_font) != CAIRO_STATUS_SUCCESS))
     {
@@ -140,6 +136,9 @@ done:
 	}
     }
 
+  _pango_cairo_font_private_scaled_font_data_destroy (cf_priv->data);
+  cf_priv->data = NULL;
+
   return cf_priv->scaled_font;
 }
 
@@ -151,7 +150,8 @@ done:
  * The scaled font can be referenced and kept using
  * cairo_scaled_font_reference().
  *
- * Return value: the #cairo_scaled_font_t used by @font
+ * Return value: the #cairo_scaled_font_t used by @font,
+ *               or %NULL if @font is %NULL.
  *
  * Since: 1.18
  **/
@@ -160,15 +160,8 @@ pango_cairo_font_get_scaled_font (PangoCairoFont *cfont)
 {
   PangoCairoFontPrivate *cf_priv;
 
-  if (G_UNLIKELY (!PANGO_IS_CAIRO_FONT (cfont)))
-    {
-      if (!_pango_cairo_warning_history.font_get_scaled_font)
-	{
-	  _pango_cairo_warning_history.font_get_scaled_font = TRUE;
-	  g_warning ("pango_cairo_font_get_scaled_font called with bad font, expect ugly output");
-	}
-      return NULL;
-    }
+  if (G_UNLIKELY (!cfont))
+    return NULL;
 
   cf_priv = PANGO_CAIRO_FONT_PRIVATE (cfont);
 
@@ -327,6 +320,7 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
   gravity = pango_font_description_get_gravity (desc);
 
   cairo_scaled_font_get_ctm (scaled_font, &cairo_ctm);
+  cairo_scaled_font_get_font_options (scaled_font, font_options);
   /* I started adding support for vertical hexboxes here, but it's too much
    * work.  Easier to do with cairo user fonts and vertical writing mode
    * support in cairo.
@@ -378,15 +372,15 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
     pango_font_description_set_family_static (desc, "monospace");
 
     rows = 2;
-    mini_size = size / 2.4;
+    mini_size = size / 2.2;
     if (is_hinted)
       {
 	mini_size = HINT_Y (mini_size);
 
-	if (mini_size < 5.0)
+	if (mini_size < 6.0)
 	  {
 	    rows = 1;
-	    mini_size = MIN (MAX (size - 1, 0), 5.0);
+	    mini_size = MIN (MAX (size - 1, 0), 6.0);
 	  }
       }
 
@@ -408,7 +402,7 @@ _pango_cairo_font_private_get_hex_box_info (PangoCairoFontPrivate *cf_priv)
   cairo_font_options_destroy (font_options);
 
 
-  scaled_mini_font = pango_cairo_font_get_scaled_font (mini_font);
+  scaled_mini_font = pango_cairo_font_get_scaled_font ((PangoCairoFont *) mini_font);
 
   for (i = 0 ; i < 16 ; i++)
     {
@@ -576,6 +570,7 @@ _pango_cairo_font_private_get_glyph_extents_missing (PangoCairoFontPrivate *cf_p
 						     PangoRectangle        *logical_rect)
 {
   PangoCairoFontHexBoxInfo *hbi;
+  gunichar ch;
   gint rows, cols;
 
   hbi = _pango_cairo_font_private_get_hex_box_info (cf_priv);
@@ -585,8 +580,13 @@ _pango_cairo_font_private_get_glyph_extents_missing (PangoCairoFontPrivate *cf_p
       return;
     }
 
+  ch = glyph & ~PANGO_GLYPH_UNKNOWN_FLAG;
+
   rows = hbi->rows;
-  cols = ((glyph & ~PANGO_GLYPH_UNKNOWN_FLAG) > 0xffff ? 6 : 4) / rows;
+  if (G_UNLIKELY (glyph == PANGO_GLYPH_INVALID_INPUT || ch > 0x10FFFF))
+    cols = 1;
+  else
+    cols = ((glyph & ~PANGO_GLYPH_UNKNOWN_FLAG) > 0xffff ? 6 : 4) / rows;
 
   if (ink_rect)
     {
