@@ -158,7 +158,7 @@ static gboolean fontset_hash_key_equal (const FontsetHashKey *key_a,
 					const FontsetHashKey *key_b);
 static void     fontset_hash_key_free  (FontsetHashKey       *key);
 
-G_DEFINE_TYPE (PangoFcFontMap, pango_fc_font_map, PANGO_TYPE_FONT_MAP)
+G_DEFINE_ABSTRACT_TYPE (PangoFcFontMap, pango_fc_font_map, PANGO_TYPE_FONT_MAP)
 
 static void
 pango_fc_font_map_init (PangoFcFontMap *fcfontmap)
@@ -277,10 +277,9 @@ pango_fc_coverage_key_equal (PangoFcCoverageKey *key1,
 
 struct _FontsetHashKey {
   PangoFcFontMap *fontmap;
-  PangoMatrix matrix;
   PangoLanguage *language;
   PangoFontDescription *desc;
-  int size;			/* scaled via the current DPI */
+  int scaled_size;
   gpointer context_key;
 };
 
@@ -318,13 +317,9 @@ static gboolean
 fontset_hash_key_equal (const FontsetHashKey *key_a,
 			const FontsetHashKey *key_b)
 {
-  if (key_a->size == key_b->size &&
+  if (key_a->scaled_size == key_b->scaled_size &&
       pango_font_description_equal (key_a->desc, key_b->desc) &&
-      key_a->language == key_b->language &&
-      key_a->matrix.xx == key_b->matrix.xx &&
-      key_a->matrix.xy == key_b->matrix.xy &&
-      key_a->matrix.yx == key_b->matrix.yx &&
-      key_a->matrix.yy == key_b->matrix.yy)
+      key_a->language == key_b->language)
     {
       if (key_a->context_key)
 	return PANGO_FC_FONT_MAP_GET_CLASS (key_a->fontmap)->context_key_equal (key_a->fontmap,
@@ -342,11 +337,6 @@ fontset_hash_key_hash (const FontsetHashKey *key)
 {
     guint32 hash = FNV1_32_INIT;
 
-    /* We do a bytewise hash on the context matrix */
-    hash = hash_bytes_fnv ((unsigned char *)(&key->matrix),
-			   sizeof(double) * 4,
-			   hash);
-
     if (key->context_key)
       hash ^= PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_hash (key->fontmap,
 									    key->context_key);
@@ -354,7 +344,7 @@ fontset_hash_key_hash (const FontsetHashKey *key)
     /* 1237 is just an abitrary prime */
     return (hash ^
 	    GPOINTER_TO_UINT (key->language) ^
-	    (key->size * 1237) ^
+	    (key->scaled_size * 1237) ^
 	    pango_font_description_hash (key->desc));
 }
 
@@ -376,10 +366,9 @@ fontset_hash_key_copy (FontsetHashKey *old)
   FontsetHashKey *key = g_slice_new (FontsetHashKey);
 
   key->fontmap = old->fontmap;
-  key->matrix = old->matrix;
   key->language = old->language;
   key->desc = pango_font_description_copy (old->desc);
-  key->size = old->size;
+  key->scaled_size = old->scaled_size;
   if (old->context_key)
     key->context_key = PANGO_FC_FONT_MAP_GET_CLASS (key->fontmap)->context_key_copy (key->fontmap,
 										     old->context_key);
@@ -554,7 +543,7 @@ font_hash_key_for_context (PangoFcFontMap *fcfontmap,
     key->context_key = NULL;
 }
 
-/* Add a mapping from xfont->font_pattern to xfont */
+/* Add a mapping from fcfont->font_pattern to fcfont */
 static void
 pango_fc_font_map_add (PangoFcFontMap *fcfontmap,
 		       PangoContext   *context,
@@ -592,7 +581,7 @@ pango_fc_font_map_lookup (PangoFcFontMap *fcfontmap,
   return g_hash_table_lookup (priv->font_hash, &key);
 }
 
-/* Remove mapping from xfont->font_pattern to xfont */
+/* Remove mapping from fcfont->font_pattern to fcfont */
 void
 _pango_fc_font_map_remove (PangoFcFontMap *fcfontmap,
 			   PangoFcFont    *fcfont)
@@ -834,6 +823,7 @@ pango_fc_make_pattern (const  PangoFontDescription *description,
    * dpi values.
    */
   pattern = FcPatternBuild (NULL,
+			    PANGO_FC_VERSION, FcTypeInteger, pango_version(),
 			    FC_WEIGHT, FcTypeInteger, weight,
 			    FC_SLANT,  FcTypeInteger, slant,
 #ifdef FC_WIDTH
@@ -1022,20 +1012,20 @@ pango_fc_font_map_get_resolution (PangoFcFontMap *fcfontmap,
 }
 
 static int
-get_unscaled_size (PangoFcFontMap             *fcfontmap,
-		   PangoContext               *context,
-		   const PangoFontDescription *desc)
+get_scaled_size (PangoFcFontMap             *fcfontmap,
+		 PangoContext               *context,
+		 const PangoFontDescription *desc)
 {
-  int size = pango_font_description_get_size (desc);
+  double size = pango_font_description_get_size (desc);
 
-  if (pango_font_description_get_size_is_absolute (desc))
-    return size;
-  else
+  if (!pango_font_description_get_size_is_absolute (desc))
     {
       double dpi = pango_fc_font_map_get_resolution (fcfontmap, context);
 
-      return (int)(0.5 + size * dpi / 72.);
+      size = size * dpi / 72.;
     }
+
+  return .5 + pango_matrix_get_font_scale_factor (pango_context_get_matrix (context)) * size;
 }
 
 static PangoFcPatternSet *
@@ -1051,17 +1041,17 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
   int f;
   PangoFcPatternSet *patterns;
   FcFontSet *font_patterns;
+  PangoMatrix matrix;
   FontsetHashKey key;
 
   if (!language && context)
     language = pango_context_get_language (context);
 
   key.fontmap = fcfontmap;
-  get_context_matrix (context, &key.matrix);
+  key.scaled_size = get_scaled_size (fcfontmap, context, desc);
   key.language = language;
   key.desc = pango_font_description_copy_static (desc);
   pango_font_description_unset_fields (key.desc, PANGO_FONT_MASK_SIZE);
-  key.size = get_unscaled_size (fcfontmap, context, desc);
 
   if (PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get)
     key.context_key = (gpointer)PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->context_key_get (fcfontmap, context);
@@ -1072,10 +1062,8 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
 
   if (patterns == NULL)
     {
-      double scale_factor = pango_matrix_get_font_scale_factor (&key.matrix);
-      double scaled_size = key.size * scale_factor / PANGO_SCALE;
       pattern = pango_fc_make_pattern (desc, language,
-				       scaled_size,
+				       key.scaled_size / 1024.,
 				       pango_fc_font_map_get_resolution (fcfontmap, context));
 
       pango_fc_default_substitute (fcfontmap, context, pattern);
@@ -1084,18 +1072,14 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
 
       if (!font_patterns)
 	{
-	  g_printerr ("No fonts found; this probably means that the fontconfig\n"
+	  g_critical ("No fonts found:\n"
+		      "This probably means that the fontconfig\n"
 		      "library is not correctly configured. You may need to\n"
 		      "edit the fonts.conf configuration file. More information\n"
 		      "about fontconfig can be found in the fontconfig(3) manual\n"
-		      "page and on http://fontconfig.org\n");
+		      "page and on http://fontconfig.org");
 
-	  /* There is no point in proceeding; we'll just get a segfault later
-	   * on, and a bunch more possibly confusing error messages in between.
-	   */
-
-	  /* return NULL; */
-	  exit (1);
+	  font_patterns = FcFontSetCreate ();
 	}
 
       patterns = g_slice_new (PangoFcPatternSet);
@@ -1124,7 +1108,7 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
 
       FcPatternDestroy (pattern);
 
-      FcFontSetSortDestroy (font_patterns);
+      FcFontSetDestroy (font_patterns);
 
       patterns->key = fontset_hash_key_copy (&key);
       g_hash_table_insert (priv->fontset_hash,
@@ -1373,7 +1357,6 @@ _pango_fc_font_map_get_coverage (PangoFcFontMap *fcfontmap,
  *
  * Since: 1.6
  **/
-
 PangoCoverage  *
 _pango_fc_font_map_fc_to_coverage (FcCharSet *charset)
 {
@@ -1396,7 +1379,7 @@ _pango_fc_font_map_fc_to_coverage (FcCharSet *charset)
 	  FcChar32  bits = map[i];
 	  FcChar32  base = ucs4 + i * 32;
 	  int b = 0;
-	  bits = map[i];
+
 	  while (bits)
 	    {
 	      if (bits & 1)
